#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : ann4qa
# @Time         : 2023/4/24 18:10
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :


from meutils.pipe import *
from meutils.np_utils import cosine_topk
from chatllm.applications import ChatBase

from sentence_transformers import SentenceTransformer


class ChatANN(ChatBase):

    def __init__(self, backend='in_memory', encode_model="moka-ai/m3e-small", **kwargs):
        """
        :param backend:
            'in_memory' # todo: 支持更多后端
        :param encode_model:
            "nghuyong/ernie-3.0-nano-zh"
            "shibing624/text2vec-base-chinese"
            "GanymedeNil/text2vec-large-chinese"
        :param kwargs:
        """
        super().__init__(**kwargs)
        self.backend = backend
        self.encode = SentenceTransformer(encode_model).encode  # 加缓存，可重新set

        # create index
        self.index = None

        # 召回结果df
        self.recall = pd.DataFrame({'id': [], 'text': [], 'score': []})

    def qa(self, query, topk=3, threshold=0.66, **kwargs):
        df = self.find(query, topk, threshold)
        if len(df):
            knowledge_base = '\n'.join(df.text)
            return self._qa(query, knowledge_base, **kwargs)
        logger.error('召回内容为空!!!')

    def find(self, query, topk=5, threshold=0.66):  # 返回df
        v = self.encode([query])  # ndim=2

        if self.backend == 'in_memory':
            idxs, scores = cosine_topk(v, np.array(self.index.embedding.tolist()), topk)

            self.recall = (
                self.index.iloc[idxs, :]
                .assign(score=scores)
                .query(f'score > {threshold}')
            )

        return self.recall

    def create_index(self, texts):  # todo：增加 encode_model参数
        embeddings = self.encode(texts, show_progress_bar=True)
        if self.backend == 'in_memory':
            self.index = pd.DataFrame({'text': texts, 'embedding': embeddings.tolist()})

        return self.index


if __name__ == '__main__':

    qa = ChatANN(encode_model="nghuyong/ernie-3.0-nano-zh")
    qa.load_llm(model_name_or_path="/CHAT_MODEL/chatglm-6b")
    qa.create_index(['周杰伦'] * 10)

    for i in qa(query='有几个周杰伦'):
        print(i, end='')
    print(qa.recall)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : to_chat
# @Time         : 2024/1/22 12:09
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  : todo: 聊天即画图

"""
chat-dall-e-3

"""
from meutils.pipe import *
from meutils.decorators.retry import retrying

from openai import OpenAI, AsyncOpenAI
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk

from chatllm.schemas.openai_types import ChatCompletionRequest, ImageRequest
from chatllm.schemas.openai_types import chat_completion_error, chat_completion_chunk_error, chat_completion_chunk


class Completions(object):
    def __init__(self, api_key: Optional[str] = None, base_url: Optional[str] = None, **kwargs):
        # 需要兜底的模型
        params = dict(
            api_key=api_key,
            base_url=base_url or 'https://api.chatllm.vip/v1',
            http_client=httpx.Client(follow_redirects=True)
        )
        # 如果这一步就报错呢
        # self.client = OpenAI(**params)  # todo: 异步
        self.client = AsyncOpenAI(**params)  # todo: 异步

    def create(self, request: ChatCompletionRequest, **kwargs):
        prompt = request.messages[-1]['content']  # 仅支持单轮

        if request.model.startswith(('chat-dall-e', 'dall-e')):  # ToDo：触发词 prompt.startswith(('画',))
            image_request = ImageRequest(prompt=prompt)
            return self.openai_image_create(image_request)

        elif request.model.startswith(('chat-sd',)):

            return self.sd_image_create()
        elif request.model.startswith(('chat-mj',)):
            return self.mj_image_create()

        return self.client.completions.create(**request.model_dump())  # 聊天兜底

    # Queue().put()

    # @retrying
    def openai_image_create(self, image_request: ImageRequest):
        """
            class ImagesResponse(BaseModel):
                created: int

                data: List[Image]
        :param image_request:
        :return:
        """
        # 前置
        chunk = chat_completion_chunk.model_copy(deep=True)
        for i in self.progress_bar():
            chunk.choices[0].delta.content = i
            yield chunk

        response = await self.client.images.generate(**image_request.model_dump())
        # return response

        for i, image in enumerate(response.data):
            image_url = image.url
            # b64_json = image.b64_json
            revised_prompt = image.revised_prompt
            chunk.choices[0].delta.content = f"""
            ![{revised_prompt}]({image_url})
            """
            yield chunk

    def sd_image_create(self):
        """gpt增强提示词"""
        pass

    def mj_image_create(self):
        """gpt增强提示词"""
        pass

    def progress_bar(self, title="**🔥ChatfireProgress** task-id", interval=1):
        yield title

        for i in range(10, 100, 10):
            time.sleep(interval)
            # yield f"![Progress](https://progress-bar.dev/{i})"
            yield f"![Progress](https://api.chatllm.vip/minio/bar/{i}.svg)"
            # "[下载图片](https://ai-mj-images.oss-cn-hangzhou.aliyuncs.com/2024-01-24/rdojtqkahmdkgmqspyoykfczqsnkkrrr.png) "


if __name__ == '__main__':

    def main(self, image_request: ImageRequest):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.openai_image_create(image_request))


    r = Completions().create(
        ChatCompletionRequest(model='chat-dall-e-3', messages=[{'role': 'user', 'content': '画个猫'}])
    )

    for i in r:
        print(i)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : apps
# @Time         : 2023/12/14 19:30
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :


from meutils.serving.fastapi import App
from chatllm.llmchain import init_cache
from chatllm.api.routers import all_chat_completions, embeddings

init_cache()

app = App()

app.include_router(embeddings.router, '/v1')
app.include_router(all_chat_completions.router, '/v1')

if __name__ == '__main__':
    app.run(port=39999)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : openai_embeddings
# @Time         : 2024/1/11 09:03
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from meutils.pipe import *
from meutils.notice.feishu import send_message

from openai import OpenAI

send_message = partial(
    send_message,
    title="Embedding主备",
    url="https://open.feishu.cn/open-apis/bot/v2/hook/e2f5c8eb-4421-4a0b-88ea-e2d9441990f2"
)


@lru_cache()
class Embeddings(object):
    def __init__(self, api_key: Optional[str] = None, base_url: Optional[str] = None, **kwargs):
        # 需要兜底的模型
        params = dict(
            api_key=api_key or 'sk-...',
            base_url=base_url or 'https://api.githubcopilot.com',
            default_headers={'Editor-Version': 'vscode/1.85.1'},
        )
        self.client = OpenAI(**params)  # todo: 异步

        self.backup_client = OpenAI(
            api_key=os.getenv("BACKUP_API_KEY"),
            base_url=os.getenv("OPENAI_BASE_URL", "https://api.chatllm.vip/v1")
        )

    def create(self, **data) -> Any:
        # data = {key: data.get(key) for key in completion_keys if key in data}  # 去除多余key

        creates = [self.client.embeddings.create, self._backup_create, ]

        for i, _create in enumerate(creates):
            try:
                response = _create(**data)  # 尝试执行操作

                # break  # 如果操作成功，则跳出循环
                response.usage.total_tokens = 50000
                response.usage.prompt_tokens = 50000
                response.usage.completion_tokens = 100000
                return response

            except Exception as e:  # 走兜底
                _ = f"Client {i} failed: {e}"
                send_message(_)
                logging.error(_)

        return self._handle_error(data, "未知错误，请联系管理员")

    def _backup_create(self, **data):
        """恢复模型名"""
        backup_data = data.copy()
        backup_data['model'] = f"backup-{data['model']}"  # "backup-text-embedding-ada-002"
        backup_response = self.backup_client.embeddings.create(**backup_data)

        send_message(f"入参：{data}")  # 兜底监控

        return backup_response

    def _handle_error(self, data: Dict[str, Any], error: Union[str, Exception]) -> Any:
        """
        Handle errors and return an appropriate response.
        """
        return f"{data}: {error}"


if __name__ == '__main__':
    from chatllm.llmchain.completions.github_copilot import Completions

    api_key = Completions.get_access_token('xx')
    print(Embeddings(api_key=api_key).create(input=['傻逼'] * 5, model='text-embedding-ada-002'))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : copilot
# @Time         : 2023/12/6 13:14
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  : 不准白嫖 必须 star, todo: 兜底设计、gpt/图片
# https://github.com/CaoYunzhou/cocopilot-gpt/blob/main/main.py
# todo: chatfire sdk

import openai
from meutils.pipe import *
from meutils.cache_utils import ttl_cache
from meutils.decorators.retry import retrying
from meutils.queues.uniform_queue import UniformQueue

from openai.types.chat.chat_completion import ChatCompletion
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk

from chatllm.llmchain.completions import openai_completions
from chatllm.schemas.openai_api_protocol import ChatCompletionRequest, UsageInfo


class Completions(object):
    def __init__(self, **client_params):
        self.api_key = client_params.get('api_key')
        self.access_token = self.get_access_token(api_key=self.api_key)
        self.completions = openai_completions.Completions(api_key=self.access_token)

    def create(self, request: ChatCompletionRequest, **kwargs):
        # logger.debug(request)

        request.model = request.model.startswith('gpt-4') and 'gpt-4' or 'gpt-3.5-turbo'

        if request.stream:
            stream = self.completions.create(request)
            # return stream

            interval = 0.05 if "gpt-4" in request.model else 0.01
            return UniformQueue(stream).consumer(interval=interval, break_fn=self.break_fn)

        else:
            return self.completions.create(request)

    def create_sse(self, request: ChatCompletionRequest):
        response = self.create(request)
        if request.stream:
            from sse_starlette import EventSourceResponse
            generator = (chunk.model_dump_json() for chunk in response)
            return EventSourceResponse(generator, ping=10000)
        return response

    @staticmethod
    def get_access_token(api_key):  # embedding 依赖于此
        token = Completions._get_access_token(api_key=api_key)
        # expires_at = int(token.split('exp=')[1][:10])  # tid=7fdaed051cf26939e7cc11e4aed37893;exp=1705749197
        if token and int(token.split('exp=')[1][:10]) < time.time():  # 过期刷新
            token = Completions._get_access_token(api_key, dynamic_param=time.time() // 180)  # 缓存3分钟
        return token

    @staticmethod
    @retrying
    @ttl_cache(ttl=1200)  # todo: 清除缓存
    def _get_access_token(api_key: Optional[str] = None, dynamic_param=None):

        api_key = api_key or os.getenv("GITHUB_COPILOT_TOKEN")
        assert api_key
        host = 'api.github.com'
        if api_key.startswith('ccu_'):  # api_key.startswith('ghu_')
            host = 'api.cocopilot.org'
        elif api_key.startswith('net|'):  # 新车
            host = 'api.cocopilot.net'
            api_key = api_key.strip('net|')

        headers = {
            'Host': host,
            # 'authorization': f'Bearer {api_key}',
            'authorization': f'token {api_key}',

            'Editor-Version': 'vscode/1.85.1',
            'Editor-Plugin-Version': 'copilot-chat/0.11.1',
            'User-Agent': 'GitHubCopilotChat/0.11.1',
            'Accept': '*/*',
            # "Accept-Encoding": "gzip, deflate, br"
        }

        url = f"https://{headers.get('Host')}/copilot_internal/v2/token"
        response = requests.get(url, headers=headers)  # todo: httpx
        resp = response.json()

        # resp['api_key'] = api_key
        # logger.debug(resp)

        return resp.get('token')

    @staticmethod
    def break_fn(line: ChatCompletionChunk):
        return line.choices and line.choices[0].finish_reason

    @classmethod
    def chat(cls, data: dict):  # TEST
        """
            Completions.chat(data)
        """

        # todo: 统一请求体
        with timer('聊天测试'):
            _ = cls().create(ChatCompletionRequest.model_validate(data))

            print(f'{"-" * 88}\n')
            if isinstance(_, Generator) or isinstance(_, openai.Stream):
                for i in _:
                    content = i.choices[0].delta.content
                    if content:
                        print(content, end='')
            else:
                print(_.choices[0].message.content)
            print(f'\n\n{"-" * 88}')


if __name__ == '__main__':
    # 触发风控
    s = """
    Question:已知节点类型只有六种：原因分析、排故方法、故障时间、故障现象、故障装备单位、训练地点，现在我给你一个问题，你需要根据这个句子来推理出这个问题的答案在哪个节点类型中，问题是”管道细长、阻力太大时的轴向柱塞泵故障如何解决？“,输出格式形为：["节点类型1"], ["节点类型2"], …。除了这个列表以外请不要输出别的多余的话。
['排故方法']

Question:已知节点类型只有六种：原因分析、排故方法、故障时间、故障现象、故障装备单位、训练地点，现在我给你一个问题，你需要根据这个句子来推理出这个问题的答案在哪个节点类型中，问题是”转向缸出现爬行现象，但是压力表却忽高忽低，相对应的解决方案是？“输出格式形为：["节点类型1"], ["节点类型2"], …。除了这个列表以外请不要输出别的多余的话。
['原因分析']、['排故方法']

Question:已知节点类型只有六种：原因分析、排故方法、故障时间、故障现象、故障装备单位、训练地点，现在我给你一个问题，你需要根据这个句子来推理出这个问题的答案在哪个节点类型中，问题是”在模拟训练场A，轴向柱塞马达出现过什么故障？“输出格式形为：["节点类型1"], ["节点类型2"], …。除了这个列表以外请不要输出别的多余的话。

['故障现象']

已知节点类型只有六种：原因分析、排故方法、故障时间、故障现象、故障装备单位、训练地点，现在我给你一个问题，你需要根据这个句子来推理出这个问题的答案在哪个节点类型中，问题是”密封圈挤出间隙的解决方法是什么？“。输出格式形为：["节点类型1"], ["节点类型2"], …。除了这个列表以外请不要输出别的多余的话。
    """

    # s = "1+1"
    # s = '树上9只鸟，打掉1只，还剩几只'
    # s = '讲个故事'

    data = {
        'model': 'gpt-4',
        'messages': [
            {'role': 'system', 'content': 'You are a helpful assistant.'},
            {'role': 'user', 'content': s}
        ],
        'stream': True
    }

    Completions.chat(data)
    # data['stream'] = True
    # Completions.chat(data)

    # async def main():
    #     _ = await Completions().acreate(**data)
    #
    #     content = ''
    #     for i in _:
    #         content += i.choices[0].delta.content
    #     return content
    #
    #
    # print(arun(main()))

    # with timer('异步'):
    #     print([Completions().acreate(**data) for _ in range(10)] | xAsyncio)

    # data = {
    #     'model': 'gpt-xxx',
    #     'messages': [{'role': 'user', 'content': '讲个故事。 要足够长，这对我很重要。'}],
    #     'stream': False,
    #     # 'max_tokens': 16000
    # }
    # data = {
    #     'model': 'gpt-4',
    #     'messages': '树上9只鸟，打掉1只，还剩几只',  # [{'role': 'user', 'content': '树上9只鸟，打掉1只，还剩几只'}],
    #     'stream': False,
    #     'temperature': 0,
    #     # 'max_tokens': 32000
    # }
    #
    # for i in tqdm(range(1000)):
    #     _ = Completions().create(**data)
    #     print(_.choices[0].message.content)
    #     break
